import logging

from Function import Function

# https://github.com/intel/linux-sgx/blob/master/SampleCode/SampleEnclave/Enclave/Edger8rSyntax/Pointers.edl

class EdlFileWriter:
    
    def __init__ (self, filePath):
        self.trustedFunctions = {}
        self.untrustedFunctions = {}
        self.types = [] #list of code representing struct union and enum
        self.filePath = filePath
        self.define = ""
        self.include = ""
    def addDefines(self, define: str):
        """Add #define at the start of the edl file"""
        self.define = define

    def addIncludes(self):
        """Add #include at the start of the file, used especially for enclave_ra_init"""
        self.include += ""

    def addTrustedFunction(self, function: Function):
        """ Add a new trusted function if it is not already present in untrusted section """
        if function.functionName in self.untrustedFunctions:  
            raise Exception("Error #E001: the function trusted" + function.functionName + "is already present in untrusted function")
        
        self.trustedFunctions[function.functionName] = function

    def addUntrustedFunction(self, function: Function):
        """ Add a new Untrusted function if it is not already present in trusted section """
        if function.functionName in self.trustedFunctions:  
            raise Exception("Error #E002: the function untrusted" + function.functionName + "is already present in trusted function")
        
        self.untrustedFunctions[function.functionName] = function

        
    def setTypeDefinitions(self, typeDefinitionCode: []):
        self.types = typeDefinitionCode 

    def writeEdlFile(self,remoteAttestation,enclave_to_be_attested: bool):
        """ Try to create a new file, override the file if already exists """
        with open(self.filePath, "w") as file: 
            logging.info("Writing file: " + self.filePath )
            file.write(self.include)
            file.write(self.getContent(remoteAttestation,enclave_to_be_attested))
            
    def getContent(self,remoteAttestation,enclave_to_be_attested):
        fileContent =  f"""
/*
* This code as been generated automatically by Python Intel SGX converter script
*/

{self.define}

enclave {{
    include "sgx_trts.h"
    include "sgx_utils.h"
    include "sgx_tkey_exchange.h"\n """
        if remoteAttestation == 2 or (remoteAttestation == 1 and enclave_to_be_attested is True) :
            fileContent = fileContent + f"""
    from "sgx_tkey_exchange.edl" import *;
    from "sgx_tstdc.edl" import *;
"""
        for typedef in self.types:
            fileContent +=  "        " + typedef.replace("\n", "\n        ")  + "\n\n"

        fileContent += """
        trusted {\n"""

        for f in self.trustedFunctions:
            trusted: Function = self.trustedFunctions[f]
            fileContent += "            public " + trusted.returnType + " " + trusted.functionName + trusted.getSgxForEdl() + ';\n'

        #if remote attestation is activated, write the functions for the process
        if remoteAttestation==1 and enclave_to_be_attested is True :
            fileContent+=r"""
            public sgx_status_t get_report([out] sgx_report_t *report,[in] sgx_target_info_t *target_info);

            public sgx_status_t enclave_ra_init(sgx_ec256_public_t key, int b_pse,[out] sgx_ra_context_t *ctx, [out] sgx_status_t *pse_status);

            public sgx_status_t enclave_ra_init_def(int b_pse,[out] sgx_ra_context_t *ctx, [out] sgx_status_t *pse_status);

            public sgx_status_t enclave_ra_get_key_hash([out] sgx_status_t *get_keys_status, sgx_ra_context_t ctx,sgx_ra_key_type_t type, [out] sgx_sha256_hash_t *hash);

            public sgx_status_t enclave_ra_close(sgx_ra_context_t ctx);
            """


        fileContent += """
        };

        untrusted {\n"""

        for f in self.untrustedFunctions:
            untrusted: Function = self.untrustedFunctions[f]
            fileContent += "             " + untrusted.returnType + " " + untrusted.functionName + untrusted.getSgxForEdl() + ';\n'

        fileContent += """
        };
};
"""
        return fileContent
