import regex
from Function import Function
from gccInterface import removeCommentsFromSourceCode
import re
import logging


class cFileReaderWriter(object):
    """ Read and close a file then stores its content.  
    Raise IOError if the file is not found
    """

    def __init__(self, filePath, new=False):
        self.filePath = filePath        
        self.functions = {}
        self.structs = {}
        self.enums = {}
        self.unions = {}
        self.includes = {}
        self.defines = []
        self.contents = ""

        if not new:
            self.contents = removeCommentsFromSourceCode(self.filePath)  # remove comments

            # parsing file
            self.__parseFunctions__()
            self.__parseStructs__()
            self.__parseEnums__()
            self.__parseUnions__()
            self.__parseInclude__()
            self.__parseDefine__()

            logging.info(f"""There are {len(self.functions)} functions: {list(self.functions)}""")
            logging.debug(f"""There are {len(self.structs)} structs: {list(self.structs)}""")
            logging.debug(f"""There are {len(self.enums)} enums: {list(self.enums)}""")
            logging.debug(f"""There are {len(self.unions)} unions: {list(self.unions)}""")
            logging.debug(f"""There are {len(self.includes)} include: {list(self.includes)}""")
            logging.debug(f"""There are {len(self.defines)} define: {self.defines}""")

    def __parseFunctions__(self): 
        logging.debug("parsing functions")
        
        groups = list(regex.findall(r"(^(static)?[\s]*([\w]+)[\s]+([\w]+)[\s]*(\([^()]*\))[\s]*(\{(?:[^{}]+|(?6))*\}))", self.contents, flags=regex.MULTILINE))

        for function in groups:            
            entireCode, static, returnType, functionName, arguments, implementationCode = function 
            self.functions[functionName] = Function(static, returnType, functionName, arguments, implementationCode, entireCode, self.filePath)

    def __parseStructs__(self):
        logging.debug("parsing structs")

        match = regex.findall(r"(^[ ]*struct[\s]*([\w]*)[\s]*{[^}]*}[\s]*;)", self.contents, flags=regex.MULTILINE)
        for struct in match:
            structImpl, structName = struct
            self.structs[structName] = structImpl

    def __parseEnums__(self):
        logging.debug("parsing enums")

        match = regex.findall(r"(^[ ]*enum[\s]*([\w]*)[\s]*{[^}]*}[\s]*;)", self.contents, flags=regex.MULTILINE)
        for enum in match:
            enumImpl, enumName = enum
            self.enums[enumName] = enumImpl

    def __parseUnions__(self):
        logging.debug("parsing unions")

        match = regex.findall(r"(^[ ]*union[\s]*([\w]*)[\s]*{[^}]*}[\s]*;)", self.contents, flags=regex.MULTILINE)
        for union in match:
            unionImpl, unionName = union
            self.unions[unionName] = unionImpl

    def __parseDefine__(self):
        logging.debug("parsing define")

        match = regex.findall(r"(\#define[^\n]*)", self.contents, flags=regex.MULTILINE)
        for define in match:
            if 'sgx_' not in define:  # only define not macro of the framework
                self.defines.append(define)

    def getSGXCopyMoveCode(self, enclave_iterator):
        """this method parse the code inside the copy_start copy_end, move_start move_end options.
        Removes the pragma annotation and remove the code inside the move option"""
        logging.debug("parsing code to be copied and moved")

        code = ""

        copies = regex.findall(r"(#[ ]*pragma[ ]*copy_start_E[\d]+)([!-~\s]*?)(#[ ]*pragma[ ]*copy_end_E[\d]+)", self.contents, flags=regex.MULTILINE)
        
        for copy in copies:
            pragma_start, codeToCopy, pragma_end = copy
            enclave_id = pragma_start.split("_")[2][1]
            if enclave_id == str(enclave_iterator):
                code += codeToCopy + "\n"
                self.contents = self.contents.replace(pragma_start, "").replace(pragma_end, "")

        moves = regex.findall(r"(#[ ]*pragma[ ]*move_start_E[\d]+([!-~\s]*?)#[ ]*pragma[ ]*move_end_E[\d]+)", self.contents, flags=regex.MULTILINE)

        for move in moves:
            all, codeToMove = move
            enclave_id = all.split("_")[2][1]
            # the standard string is move_start_E2\n.....  we split in {'move','start','E2..'},
            # take 'E2...' as split("..")[2] and then the digit '2'
            if enclave_id == str(enclave_iterator):
                code += codeToMove + "\n"
                self.contents = self.contents.replace(all, "")

        return code

    def __findSGXFunctionsMacro__(self) -> []:
        """ find sgx framework macro, remove them and return an array of sgx definition 

        example: [{callType: callType , funcName: funcName , macroArg: macroArg}"""
        
        logging.debug("finding sgx annotation in file: " + self.filePath)
        groups_common = list(regex.findall(r"(^[ ]*#[\s]*define[\s]*sgx_(ecall|ocall)_([\w\d_]*)[\s]*(\([^()]*\)))", self.contents, flags=regex.MULTILINE))  # @TODO:decidere come gestire queste chiamate
        groups_specific = list(regex.findall
                               (r"(^[ ]*#[\s]*define[\s]*sgx_E[\d]+_(ecall|ocall)_([\w\d_]*)[\s]*(\([^()]*\)))", self.contents, flags=regex.MULTILINE))
        seen = []  # check macro duplications
        retu = []  # the array to return

        macros = []  # the list of macros
        for sgxFunc in groups_common:
            macro, callType, funcName, macroArg = sgxFunc 
            macros.append(macro)
            logging.debug("function macro arguments ok!")

            if funcName in seen:
                raise Exception("Error #A001: There is duplicates macro annotation for function: " + funcName)
            seen.append(funcName)

            retu.append({"callType": callType, "funcName": funcName, "macroArg": macroArg})

            self.contents = self.contents.replace(macro, '')  # removing the macro definition
        for sgxFunc in groups_specific:
            macro, callType, funcName, macroArg = sgxFunc
            macros.append(macro)
            logging.debug("function macro arguments ok!")

            if funcName in seen:
                raise Exception("Error #A001: There is duplicates macro annotation for function: " + funcName)
            seen.append(funcName)

            enclave = macro.split("_", 3)[1]

            retu.append({"callType": callType, "funcName": funcName, "macroArg": macroArg, "enclave": enclave})

            self.contents = self.contents.replace(macro, '')  # removing the macro definition

        logging.info("found " + str(len(macros)) + " sgx functions annotation: " + str(macros))
        return retu
    
    def getEcallOcallFunctions(self):
        """Read, check and remove the macro annotations of sgx framework\n
        return ecalls, ocalls: [name: Function], [name: Function] """
        ecallFunctions = {}
        ocallFunctions = {}

        sgxFuncList = self.__findSGXFunctionsMacro__()    # get all sgx framework definition
        logging.info("check sgx framework definition")

        for sgxFunc in sgxFuncList:
            callType, funcName, macroArg = sgxFunc["callType"], sgxFunc["funcName"], sgxFunc["macroArg"]

            if funcName not in self.functions:
                raise Exception(f"Error #A002: found macro for {funcName}, but function {funcName} is not found in file {self.filePath}")

            call: Function = self.getFunction(funcName)  # the ecall/ocall of the definition
            call.setSgxArgumentDefinition(macroArg)  # setting the macro arg of the sgx function, and check it
            lenSGX = int(sgxFunc.__len__())
            # An SGX Function that has the enclave field will have 4 fields, instead a normal SGX Func will have only 3
            if lenSGX > 3:
                call.bindFunctionAtEnclave(sgxFunc["enclave"])

            if callType == 'ecall':
                ecallFunctions[funcName] = call
            elif callType == 'ocall':
                ocallFunctions[funcName] = call 

        return ecallFunctions, ocallFunctions      

    def __parseInclude__(self):
        logging.debug("parsing includes")

        match = regex.findall(r"(^[\s]*\#[\s]*include[\s\\]*(\"[^\"]*\"|\<[^\>]*\>))$", self.contents, flags=regex.MULTILINE)
        for include in match:
            includeImpl, includePath = include
            includePath = includePath[1:-1]  # remove first and last chars
            self.includes[includePath] = includeImpl

    def getFilePath(self) -> str:
        return self.filePath

    def getAllCode(self) -> str:
        """ Return the code content """
        return self.contents

    def popFunction(self, functionName: str):
        """ Remove the function from the stored content and return its code\n
        Raise FunctionNotFount if the function is not present"""

        logging.debug('Removing function: ' + functionName)
        toRemove = self.functions.pop(functionName).entireCode
        if toRemove not in self.contents:
            raise Exception(f"Error #A003: function not found {functionName}")
        self.contents = self.contents.replace(toRemove, '')
        return toRemove

    def popStruct(self, structName):
        """ Remove the struct from the stored content and return its code\n
        Raise exception if the struct is not present"""

        logging.debug('Removing struct: ' + structName)
        toRemove = self.structs.pop(structName)
        if toRemove not in self.contents:
            raise Exception(f"Error #A004: struct not found{structName}")
        self.contents = self.contents.replace(toRemove, '')
        return toRemove

    def popEnum(self, enumName):
        """ Remove the enum from the stored content and return its code\n
        Raise exception if the enum is not present"""

        logging.debug('Removing enum: ' + enumName)
        toRemove = self.enums.pop(enumName)
        if toRemove not in self.contents:
            raise Exception(f"Error #A005: enum not found{enumName}")
        self.contents = self.contents.replace(toRemove, '')
        return toRemove

    def popUnion(self, unionName):
        """ Remove the union from the stored content and return its code\n
        Raise exception if the union is not present"""

        logging.debug('Removing union: ' + unionName)
        toRemove = self.unions.pop(unionName)
        if toRemove not in self.contents:
            raise Exception(f"Error #A006: union not found{unionName}")
        self.contents = self.contents.replace(toRemove, '')
        return toRemove

    def getFunction(self, functionName: str) -> Function:
        """ Return the Function object representing the function named functionName"""
        return self.functions.get(functionName)

    def getStruct(self, structName) -> str:
        """ Return the code of the struct """
        return self.structs.get(structName)

    def getEnum(self, enumName):
        """ Return the code of the enum """
        return self.enums.get(enumName)

    def getUnion(self, unionName):
        """ Return the code of the union """
        return self.unions.get(unionName)

    def getDefines(self):
        """ Return the code of the union """
        return self.defines

    def saveFile(self):
        f = open(self.filePath, "w")
        f.write(self.getAllCode())
        f.close()

    def getAllFunctions(self) -> dict:
        return self.functions 

    def popAllDataStructureDefinition(self):
        """ Remove all data structure defined and return the code"""
        code = []
        for s in list(self.structs.keys()):
            code.append(self.popStruct(s))
        for e in list(self.enums.keys()):
            code.append(self.popEnum(e))
        for u in list(self.unions.keys()):
            code.append(self.popUnion(u))

        return code

    def addInclude(self, filePath):
        """add a line #include at the top of the file"""
        if filePath not in self.includes:
            newInclude = f"""#include \"{filePath}\"\n"""
            self.includes[filePath] = newInclude
            self.contents = newInclude + self.contents  # add the line #include to the file

    def addFunction(self, function: Function):
        """add a function at the end of the file"""
        if function.functionName not in self.functions:
            requests = regex.findall(r"(#\s*define\s+[\w\d]+.+)", self.contents, flags=regex.MULTILINE)
            if requests :
                self.contents = self.contents.replace(requests[-1], requests[-1] + f"\n{function.getPrototype()}\n")  # we add the prototype of the function after the structs
            else:
                self.contents = function.getPrototype() + self.contents
            self.contents += "\n" + function.entireCode + "\n"
            function.filePath = self.filePath
            self.functions[function.functionName] = function

    def addStruct(self, newStruct):
        """add a struct at top of the file after the last #define"""
        if newStruct.structName not in self.structs[newStruct.structName]:
            structImpl, structName = newStruct
            self.structs[structName] = structImpl

    def replaceCode(self, a, b):
        """replace the code a with code b"""
        self.contents = self.contents.replace(a, b) 

    def replaceFunction(self, a, b):
        """Replace a function with an existing one and place it in the same position"""

        logging.debug(f"""Replacing function: {a.functionName} with {b.functionName}""")
        toReplace = self.getFunction(a.functionName).entireCode

        if toReplace not in self.contents:
            raise Exception(f"Error #A003: function not found {a.functionName}")
        self.contents = self.contents.replace(toReplace, b.entireCode)

        self.functions.pop(a.functionName, None)
        
        if b.functionName not in self.functions:
            b.filePath = self.filePath
            self.functions[b.functionName] = b

    def addCodeToTheEnd(self, code):
        """add simple code to the end of the file"""
        self.contents += "\n" + code

    def addCodeToTheStart(self, code):
        """add simple code to the start of the file"""
        self.contents = code + "\n" + self.contents

    def removePragmasFromContent(self):
        """remove the pragma comments for attestation contained in the code"""
        self.contents = re.sub(r"(#\s*pragma\s*attest_enclave_\d+\n)", '', self.contents)
        self.contents = re.sub(r"(#\s*pragma\s*SERVER_ATTESTATION\n)", '', self.contents)
        self.contents = re.sub(r"(#\s*pragma\s*move_to_server\n)", '', self.contents)
        self.contents = re.sub(r"(#\s*pragma\s*move_to_server[_E\d]+\n)", '', self.contents)

    def findFunctionRequestingClientAttestation(self):
        """ Returns the functions that asked for enclave attestation"""
        requests = regex.findall(r"(#\s*pragma\s*attest_enclave_\d+)", self.contents, flags=regex.MULTILINE)
        functions = []
        if requests:
            for f in self.functions:
                if self.functions[f].implementationCode.find("#pragma attest_enclave_") != -1:
                    # function has the pragma CA
                    functions.append(f)

        logging.info("The following functions will reserve a Remote Attestation process " + functions.__str__())
        return functions

    def findFunctionRequestingServerAttestation(self):
        requests = regex.findall(r"(#\s*pragma\s*SERVER_ATTESTATION)", self.contents,
                                 flags=regex.MULTILINE)
        functions = []
        if requests:
            for f in self.functions:
                if self.functions[f].implementationCode.find("#pragma SERVER_ATTESTATION") != -1:  #TODO:check this find (maybe with find #pragma and find SERVER_ATTESTATION
                    # function has the pragma CA
                    functions.append(f)
        return functions

    def searchForEnclaves(self, number_functions_split: int):
        """ Scan the code to find how many enclaves the app needs"""
        requests = regex.findall(r"(^\s*#[\s]*define\s*sgx_E[\d]+_(ecall|ocall))", self.contents, flags=regex.MULTILINE)
        set1 = set()
        for tuple in requests:
            set1.add(tuple[0].split("_")[1])
        if len(self.findFunctionRequestingServerAttestation()) != 0:
            set1.add("E1")
        if len(self.findFunctionRequestingClientAttestation()) != 0:
            set1.add("E1")
        if number_functions_split != 0:
            set1.add("E1")
        #return set1,len(set1)
        return set1
    #def findOrderEnclaveCalls():
    #	""" Scan the code and returns a list with the order of the enclave called. This list is used to generate the remote attestation switch cases in client_starter.cpp"""
    #	request = regexrequests = regex.findall(r"(^\s*#[\s]*define\s*sgx_E[\d]+_(ecall|ocall))", self.contents, flags=regex.MULTILINE)
        
