from ctagsInterface import c_types
import ctagsInterface
from cFileReaderWriter import cFileReaderWriter
from Function import Function
import regex
from constants import Constants
import gccInterface
from tqdm import tqdm  # https://github.com/tqdm/tqdm#installation
from EdlFileWriter import EdlFileWriter
import framaInterface
from remoteAttestationInterface import CodeRA
from codeSplittingInterface import CodeSplittingInterface

import logging
import sys
import os
import shutil
import time


start_time = time.time()

log_format = "%(levelname)-8s: %(filename)-20s.%(lineno)-3d - %(message)s"
# logging.basicConfig(stream=sys.stderr, level=logging.DEBUG, format=log_format) #CRITICAL 50, ERROR 40, WARNING 30, INFO 20, DEBUG 10, NOTSET 0
logging.basicConfig(filename="logfile.txt", filemode='w', level=logging.DEBUG,
                    format=log_format)  # CRITICAL 50, ERROR 40, WARNING 30, INFO 20, DEBUG 10, NOTSET 0


def main(filesPaths: [], src, signerSPID, iasPrimaryKey, iasSecondaryKey, iasReportFilePath, autorun, SGX_SDK,serverToConnect,MRENCLAVE,portServerCS,customReaction):

	filePath = ""
	codeSplittingActivated = False
	checkToolsAvailable()
	# legend for remote_attestation: 0 = not asked ; 1 = client attestation requested ; 2 = server attestation requested
	# ------ check part start -------------------------------------------------------------------------

	dst = "generated_trusted"
	copyRemoveAndCopyTree(filesPaths , src, dst)
	logging.info(f"Converting code from {src} to {os.getcwd()}")
	logging.info("Working in directory: " + os.getcwd())
	logging.debug("The filesPaths are " + str(filesPaths)[1:-1])
	checkFilesExistence(filesPaths)
	checkAllFileCOrH(filesPaths)
	if SGX_SDK != '~/sgxsdk':
		checkSDKDir(SGX_SDK)
		changeInstallDir(SGX_SDK)
	logging.info("checking code portability")

	ctagsFunction = ctagsInterface.getNamesOfType(filesPaths,c_types.function)  # assert all functions have unique names
	structsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.struct)  # same
	enumsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.enum)  # same
	unionsDef = ctagsInterface.getNamesOfType(filesPaths, c_types.union)  # same
	definesDef = ctagsInterface.getNamesOfType(filesPaths, c_types.define)  # same

	logging.debug("reading and parsing all files" + str(filesPaths))

	cFileReaderWriterDict = {}
	for filePath in filesPaths:  # open and parse all files
		fileReaderWriter = cFileReaderWriter(filePath)  # parse the file
		cFileReaderWriterDict[filePath] = fileReaderWriter

	numbFunctionsSplittedServer = checkPragmaCodeSplitting(cFileReaderWriterDict[filePath])
	if numbFunctionsSplittedServer > 0:
		codeSplittingActivated = True


	# -------------------------- [REMOTE ATTESTATION] Remote Attestation Implementation  ----------------------------- #
	# search for pragma client_attestation and server_attestation
	remote_attestation = CodeRA.searchForAttestation(cFileReaderWriterDict[filePath])

	# get number enclaves requested from the developer
	#enclaves_requested = cFileReaderWriterDict[filePath].searchForEnclaves(numbFunctionsSplittedServer)[1]
	enclaves_requested = cFileReaderWriterDict[filePath].searchForEnclaves(numbFunctionsSplittedServer)
	#orderEnclaves = cFileReaderWriterDict[filePath].searchForEnclaves(numbFunctionsSplittedServer)[0]
	orderEnclaves = cFileReaderWriterDict[filePath].searchForEnclaves(numbFunctionsSplittedServer)

	#print(orderEnclaves)
	logging.debug("The framework has found " + str(len(enclaves_requested)) + " enclaves requested.")

	if len(enclaves_requested) == 0 and not codeSplittingActivated:
		raise Exception("Error #B007: zero enclaves found, check the syntax of annotations. Is it correct? (sgx_E<id enclave>_ecall|ocall )")

	enclaveToAttest = search_enclaves_needing_attestation(cFileReaderWriterDict[filePath])
	# if enclaveToAttest > enclaves_requested:
	# 	raise Exception("Error #B008: the enclave asked to be requested doesn't exists.")
	# get ecallFunctions ocallFunctions allFunctions
	if codeSplittingActivated:
		#enclaves_requested = 1
		remote_attestation = 1
		if enclaveToAttest.__len__() == 0:
			enclaveToAttest.append("1")
	ecallFunctions, ocallFunctions, allFunctions = getEcallsOcallsAndAllFunc(cFileReaderWriterDict.values())
	logging.info("found ecall: " + str(Function.getNamesOfFunctionList(ecallFunctions.values())))
	logging.info("found ocall: " + str(Function.getNamesOfFunctionList(ocallFunctions.values())))

	if len(ecallFunctions) == 0  and not codeSplittingActivated and remote_attestation == 0:
		raise Exception("Error #B001: there are no ecall definition")

	if len(ctagsFunction) != len(allFunctions) and not codeSplittingActivated:  # check we read all files correctly
		raise Exception("Error #B002: something is not working with ctags and framework regex. Are there any non C11 code?")
	assertDifferentNamesOcallEcall(ecallFunctions, ocallFunctions)  # check if an ecall and ocall has different names

	# check if the definition of user type exists
	checkArgumentsTypeDefinitionInFunctions(structsDef, enumsDef, unionsDef,ecallFunctions)
	# check if the definition of user type exists
	checkArgumentsTypeDefinitionInFunctions(structsDef, enumsDef, unionsDef,ocallFunctions)
	# enclave functions are function to move into enclave
	librariesToInclude, enclaveFunctions = findAllTrustedCodeAndLibrary(ecallFunctions, ocallFunctions, allFunctions,filesPaths)
	checkSeparationOfContext(ecallFunctions, enclaveFunctions, ocallFunctions, allFunctions)

	# Find the order of the Ecalls and attest functions called in the main file
	orderFunctions = CodeRA.getOrderFunctions(cFileReaderWriterDict[filePath], ecallFunctions)

	# find the ecall needed function
	ecallNeededFunctions: list = [f for f in enclaveFunctions if f not in ecallFunctions]


	# -------------------------- [CODE SPLITTING] Code Splitting between Client and Server --------------------------- #

	# if client_attestation is asked, then we need to recognize the functions to attest
	if remote_attestation == 1:  # client attestation requested
		# setting of the client connection
		funcsNeededAttesting = cFileReaderWriterDict[filePath].findFunctionRequestingClientAttestation()
		for f in funcsNeededAttesting:
			funcMain = cFileReaderWriterDict[filePath].getFunction(f)
			all_attest_declarations = regex.findall("attest_enclave_\d+",funcMain.implementationCode,regex.MULTILINE)
			for d in all_attest_declarations:
				idEnc = d.split("_enclave_")[1]
				#CodeRA.generateClientCode(cFileReaderWriterDict[filePath], funcMain, idEnc ) ORIGINAL, UNCOMMENT IF TROUBLES
				CodeRA.generateClientCode(cFileReaderWriterDict[filePath], funcMain, idEnc,ecallFunctions)
				funcMain = cFileReaderWriterDict[filePath].getFunction(f)
		shutil.copy2("../../python/addon_files/enclave_manager_common.h","./")
		shutil.copy2("../../python/addon_files/enclave_manager_common.c", "./")
		cFileReaderWriterDict[filePath].addInclude("enclave_manager_common.h")
	elif remote_attestation == 2:  # server attestation requested
		# function that checks if the params in frameworkConfig.py are valid
		checkServerArgumentsAvailable(signerSPID, iasPrimaryKey, iasSecondaryKey, iasReportFilePath)
		cFileReaderWriterDict[filePath].addInclude("../server_material/server_interface.h")
		funcsNeededAttesting = cFileReaderWriterDict[filePath].findFunctionRequestingServerAttestation()
		for f in funcsNeededAttesting:
			funcMain = cFileReaderWriterDict[filePath].getFunction(f)
			# takes the function and insert the API for the RA in the code
			CodeRA.generateServerCode(cFileReaderWriterDict[filePath], funcMain)
		CodeRA.insertFunctionOfClientAttestation(cFileReaderWriterDict[filePath], filePath)

	if codeSplittingActivated:
		moveFunctionsToServer(cFileReaderWriterDict[filePath], filePath, remote_attestation,portServerCS,enclaveToAttest)

	cFileReaderWriterDict[filePath].removePragmasFromContent()

	enclaveFunctions = None

	# ------ creating / modifying every enclave requested ------------------------------------------------------------
	for enclave_iterator in range(1, len(enclaves_requested) + 1):
		os.mkdir("../enclave" + str(enclave_iterator))
		# creating enclave.c file
		logging.debug("generation of enclave.c for enclave" + str(enclave_iterator))
		enclaveC = cFileReaderWriter("../enclave" + str(enclave_iterator) + "/enclave.c", new=True)

		logging.debug("starting modifying the files")

		# make ocall wrappers
		for ocallName in ocallFunctions:
			ocall = allFunctions[ocallName]
			#if ocall.askedforEnclave[0][1] == str(enclave_iterator):
			OcallWrapper: Function = Function.generateOcallWrapper(ocall, enclaveC.filePath)
			enclaveC.addFunction(OcallWrapper)

		# adding ecall used function to the enclave.c
		for funName in ecallNeededFunctions:
			fun = allFunctions[funName]
			cFileReaderWriterDict[fun.filePath].popFunction(fun.functionName)  # pop the function from the original file

			enclaveC.addFunction(fun)  # add all ecall needed function in the file

		# split the ecalls in the caller(untrusted wrapper) and callee(trusted) and update the files
		for ecallName in ecallFunctions:
			if (ecallFunctions[ecallName].askedforEnclave != [] and ecallFunctions[ecallName].askedforEnclave.count(
					"E" + str(enclave_iterator)) != 0):
				ecall = allFunctions[ecallName]
				fileWhereEcallIsDef: cFileReaderWriter = cFileReaderWriterDict[
					ecall.filePath]  # get the file where function is def

				ecallTrustedFunc, ecallWrapperFunc = Function.splitFunctionToEcallWrapperAndEcallTrusted(ecall,str(enclave_iterator))  # split the function in wrapper and ecallTrusted
				fileWhereEcallIsDef.replaceFunction(ecall, ecallWrapperFunc)  # replace with the new one

				enclaveC.addCodeToTheStart(f"""#define {ecallName} {ecallTrustedFunc.functionName}""")

				enclaveC.addFunction(ecallTrustedFunc)  # add the trusted ecall function in the enclave.c file

		# copy / move code requested by user
		enclaveC.addCodeToTheStart(getCopyMoveCodes(cFileReaderWriterDict, enclave_iterator))

		# adding #define to enclave.c
		enclaveC.addCodeToTheStart(getAllDefineCode(cFileReaderWriterDict))

		# adding include to the enclave.c
		logging.debug("adding include to enclave.c for enclave" + str(enclave_iterator))

		enclaveC.addInclude(f"enclave{str(enclave_iterator)}_t.h")  # changed from caps lock to lowercase
		enclaveC.addInclude("stdlib.h")
		for library in librariesToInclude:
			enclaveC.addInclude(library)

		# adding remote attestation functions to initiate the RA process
		if remote_attestation == 1 and str(enclave_iterator) in enclaveToAttest:
			code = Constants.___remoteAttestationFunctions_Header___.replace(f"ENCLAVE[i]",f"ENCLAVE{str(enclave_iterator)}")
			enclaveC.addCodeToTheStart(code)
			enclaveC.addCodeToTheEnd(Constants.___remoteAttestationFunctions_Implementations___)

		# write the file enclave.c
		enclaveC.saveFile()

		# ---------------- edl file creation ---------------------
		logging.debug("pop all the user type code")
		typeDefinitionsCode = popAllTypedefinitionsAndGetTheCode(cFileReaderWriterDict, "enclave_manager_" + str(
			enclave_iterator) + ".h")  # get all the struct, enum, union definition code

		logging.debug("creation of enclave.edl")
		edlFile = EdlFileWriter("../enclave" + str(
			enclave_iterator) + "/enclave"+str(enclave_iterator)+".edl")  # edl file writer object (changed from caps lock to lowercase)

		logging.debug("pushing popped code to the edl file")
		edlFile.setTypeDefinitions(typeDefinitionsCode)  # add the type definition to the edl

		# define ecall definitions in edl file
		for enclaveFunctions in enclaveC.getAllFunctions().values():
			if enclaveFunctions.sgx_edl_argument_definition:  # has macro definition
				edlFile.addTrustedFunction(enclaveFunctions)

		# define ocall definitions in edl file
		for ocallName in ocallFunctions:
			ocall = allFunctions[ocallName]
			if ocall.askedforEnclave[0][1] == str(enclave_iterator):
				edlFile.addUntrustedFunction(ocall)

		# adding #define to Enclave.edl
		edlFile.addDefines(getAllDefineCode(cFileReaderWriterDict))

		if (str(enclave_iterator) in enclaveToAttest) and remote_attestation == 1:
			edlFile.writeEdlFile(remote_attestation,True)
		else:
			edlFile.writeEdlFile(remote_attestation,False)

		# ---------- creating enclave_manager.h -----------------
		Constants.writeEnclaveManager(enclave_iterator, remote_attestation,enclaveToAttest)  #other solution(2 parameters, if remote_att = 1 and enclavetoattest != 1 => remote_attest = 0

		# ----------- save all files -----------------------------
		for filePath, file in cFileReaderWriterDict.items():
			file.saveFile()

		# ----------- section for creating files needed in manual compilation in Unix ------

		if sys.platform.startswith('linux'):
			Constants.writeXMLConf("../enclave" + str(enclave_iterator) + "/enclave.config.xml", enclave_iterator)
			Constants.writeLDS("../enclave" + str(enclave_iterator) + "/enclave.lds", enclave_iterator)
			Constants.writeTestKey("../enclave" + str(enclave_iterator) + "/enclave_private_test.pem", enclave_iterator)
			if remote_attestation == 1 :
				Constants.writeMakefileRA("../Makefile", filesPaths, len(enclaves_requested))
				#if str(enclave_iterator) in enclaveToAttest:
				#	updateMakefileRA(len(enclaves_requested))

	# ----------- enabling remote attestation ----------------
	if remote_attestation == 1:  # in case of client_attestation
		shutil.copytree("../../include/materialRA/client", "../remote_attestation")
		shutil.copy2("../remote_attestation/autorun.py", "../")
		os.remove("../remote_attestation/autorun.py")

		# update of the client_starter.h file
		with open("../remote_attestation/client_starter.h", "r") as t:
			filedata = t.read()
			includes_client_starter = ""
			for id in enclaveToAttest:
				includes_client_starter += f"""#include "../application/untrusted{id}/enclave{id}_u.h" \n"""
			filedata = filedata.replace("[[UNTRUSTED_FOLDS_TO_INCLUDE]]",includes_client_starter)
			t.close()
		with open("../remote_attestation/client_starter.h", "w") as w:
			w.write(filedata)
			w.close()
		#enclaves_requested
		# update of the client_starter.cpp file for each enclave to be attested
		with open("../remote_attestation/client_starter.cpp","r") as t:
			filedata = t.read()
			enclave_ra_init = ""
			enclave_ra_init_def = ""
			enclave_ra_close = ""
			enclave_ra_get_ga = "sgx_ecall_get_ga_trusted_t sgx_ra_get_ga;"
			enc_ra_proc_msg2_t = "sgx_ecall_proc_msg2_trusted_t sgx_ra_proc_msg2_trusted;"
			enc_ra_get_msg3_t = "sgx_ecall_get_msg3_trusted_t sgx_ra_get_msg3_trusted;"
			enc_get_key_hashes = ""
			dictOrder = {}
			for i in range(0, len(orderFunctions)):
				if orderFunctions[i][0] == "A":
					enclavesCreatedBefore = 0
					idElement = orderFunctions[i][1] #we save here the X part in "AX"
					counter = 0
					while True:
						if orderFunctions[counter][1] == idElement:
							break
						enclavesCreatedBefore += 1
						counter += 1
					dictOrder[idElement] = enclavesCreatedBefore
			if enclaveToAttest.__len__()>1:

				for id in enclaveToAttest:

					enclave_ra_init += f"status = enclave{id}_enclave_ra_init(eid, &sgxrv, config.pubkey, b_pse,&ra_ctx, &pse_status);"
					if enclave_ra_init_def != "":
						enclave_ra_init_def += "\nelse "
					enclave_ra_init_def += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\nstatus= enclave{id}_enclave_ra_init_def(eid, &sgxrv, b_pse, &ra_ctx,&pse_status);"
					if enclave_ra_close != "":
						enclave_ra_close += "\nelse "
					enclave_ra_close += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\nenclave{id}_enclave_ra_close(eid, &sgxrv, ra_ctx); "
					if enclave_ra_get_ga != "sgx_ecall_get_ga_trusted_t sgx_ra_get_ga;":
						enclave_ra_get_ga += "\nelse "
					enclave_ra_get_ga += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\n sgx_ra_get_ga = enclave{id}_sgx_ra_get_ga;"
					if enc_ra_proc_msg2_t != "sgx_ecall_proc_msg2_trusted_t sgx_ra_proc_msg2_trusted;":
						enc_ra_proc_msg2_t += "\nelse "
					enc_ra_proc_msg2_t += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\nsgx_ra_proc_msg2_trusted = enclave{id}_sgx_ra_proc_msg2_trusted;"
					if enc_ra_get_msg3_t != "sgx_ecall_get_msg3_trusted_t sgx_ra_get_msg3_trusted;":
						enc_ra_get_msg3_t += "\nelse "
					enc_ra_get_msg3_t += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\nsgx_ra_get_msg3_trusted = enclave{id}_sgx_ra_get_msg3_trusted;"
					if enc_get_key_hashes != "":
						enc_get_key_hashes += "\nelse "
					enc_get_key_hashes += f"if(eid == {len(enclaves_requested)+dictOrder[str(id)]})\nstatus = enclave{id}_enclave_ra_get_key_hash(eid,&sha_status,&key_status,ra_ctx,SGX_RA_KEY_MK, &mkhash);"
			else:
				if enclaveToAttest.__len__() == 0 and codeSplittingActivated:
					enclaveToAttest = ['1']
				enclave_ra_init += f"status = enclave{enclaveToAttest[0]}_enclave_ra_init(eid, &sgxrv, config.pubkey, b_pse,&ra_ctx, &pse_status);"
				enclave_ra_init_def += f"status= enclave{enclaveToAttest[0]}_enclave_ra_init_def(eid, &sgxrv, b_pse, &ra_ctx,&pse_status);"
				enclave_ra_close += f"enclave{enclaveToAttest[0]}_enclave_ra_close(eid, &sgxrv, ra_ctx); "
				enclave_ra_get_ga += f"\nsgx_ra_get_ga = enclave{enclaveToAttest[0]}_sgx_ra_get_ga;"
				enc_ra_proc_msg2_t += f"\nsgx_ra_proc_msg2_trusted = enclave{enclaveToAttest[0]}_sgx_ra_proc_msg2_trusted;"
				enc_ra_get_msg3_t += f"sgx_ra_get_msg3_trusted = enclave{enclaveToAttest[0]}_sgx_ra_get_msg3_trusted;"
				enc_get_key_hashes += f"status = enclave{enclaveToAttest[0]}_enclave_ra_get_key_hash(eid,&sha_status,&key_status,ra_ctx,SGX_RA_KEY_MK, &mkhash);"

			filedata = filedata.replace("[ENCLAVES_RA_INIT_SWITCHES]",enclave_ra_init).replace("[ENCLAVES_RA_INIT_DEF_SWITCHES]",enclave_ra_init_def).replace("[ENCLAVES_ENCLAVE_RA_CLOSE_SWITCHES]",enclave_ra_close)
			filedata = filedata.replace("[SWITCH_CASE_RA_GET_GA]",enclave_ra_get_ga).replace("[ENCLAVES_SGX_RA_PROC_MSG2_SWITCHES]",enc_ra_proc_msg2_t).replace("[ENCLAVES_SGX_RA_GET_MSG3_TRUSTED_SWITCHES]",enc_ra_get_msg3_t)
			filedata = filedata.replace("[ENCLAVES_GET_KEY_HASH_SWITCHES]",enc_get_key_hashes)
			t.close()
		with open("../remote_attestation/client_starter.cpp","w") as writingStarter:
			writingStarter.write(filedata)
			writingStarter.close()
		CodeRA.setServerAddress(serverToConnect)
		#if len(customReaction) != 0 and (codeSplittingActivated or remote_attestation==2):
		#	includeCustomReaction(customReaction)
		if codeSplittingActivated:
			updateMakefileRA(len(enclaves_requested)) #done to not automatically add a server_cs when performing attest_enclave
	elif remote_attestation == 2:  # in case of server_attestation
		shutil.copytree("../../include/materialRA/server", "../server_material")

		# we need to insert the parameters inside the settings.h file
		CodeRA.configSPSettings(signerSPID, iasPrimaryKey, iasSecondaryKey, iasReportFilePath)
		# from ./application, copy makefile_sp from ../server_material,rename it to Makefile and delete it from server
		shutil.copy2("../server_material/Makefile_sp", "../")
		os.rename("../Makefile_sp", "../Makefile")
		os.remove("../server_material/Makefile_sp")
		CodeRA.updateMakefileServerProvider(len(enclaves_requested))
		shutil.copy2("../server_material/autorun.py", "../")
		os.remove("../server_material/autorun.py")
		substituteNameOfSourceFile(filePath)
		if MRENCLAVE == "22112021221120212211202122112021":
			allowSkipMREnclave()

	else:  # No Remote Attestation
		shutil.copy2("../../python/addon_files/autorun.py", "../")
		Constants.writeMakefile("../Makefile", filesPaths, len(enclaves_requested))
	logging.debug("Directory of remote attestation files and script to autorun the program created")
	# ---------- last operations on logfile -----------------
	ti = round(time.time() - start_time,4)
	print(ti)
	logging.info(f"Total elapsed time {round(time.time() - start_time, 4)} seconds")
	logging.shutdown()
	
	shutil.move("../../logfile.txt", "..")
	if autorun:
		os.system("cd .. && python autorun.py")
	return ti


def getCopyMoveCodes(cFileReaderWriterDict: dict, enclave_iterator) -> str:
	"""return a string representing the code to move or copy in the enclave"""

	code = ""
	for cFile in cFileReaderWriterDict.values():
		code += cFile.getSGXCopyMoveCode(enclave_iterator)

	return code


def popAllTypedefinitionsAndGetTheCode(cFileReaderWriterDict: dict, toInclude):
	"""return a list of strings representing the code popped from the files.
	Pop the definition of the struct, enum, unions from the file and include the new header"""

	typeDefinitionsCode = []
	for cFile in cFileReaderWriterDict.values():
		typeDefinitionsCode += cFile.popAllDataStructureDefinition()
		cFile.addInclude(toInclude)

	return typeDefinitionsCode


def getAllDefineCode(cFileReaderWriterDict: dict):
	"""return a string representing the code of all defines in files. (#define ...)"""

	defineCode = []
	for cFile in cFileReaderWriterDict.values():
		defineCode += cFile.getDefines()

	allDefinesString = ""

	for defineStr in defineCode:
		allDefinesString += defineStr + "\n"

	return allDefinesString


def getEcallsOcallsAndAllFunc(cFileReaderWritersList: []):
	"""open and parse all files, return 3 list of functions: ecalls, ocalls, allFuncs.
	Remove the macro annotation """

	ecallFunctions = {}  # dict of object Function
	ocallFunctions = {}  # dict of object Function
	allFunctions = {}  # dict of object Function

	for fileReaderWriter in cFileReaderWritersList:
		ecalls, ocalls = fileReaderWriter.getEcallOcallFunctions()

		ecallFunctions.update(ecalls)
		ocallFunctions.update(ocalls)
		ecalls, ocalls = None, None  # error pruning

		allFunctions.update(fileReaderWriter.getAllFunctions())

	return ecallFunctions, ocallFunctions, allFunctions


def findAllTrustedCodeAndLibrary(ecallFunctions: dict, ocallFunctions: dict, allFunctions: dict, filesPaths: []):
	"""return librariesToInclude: list of dict, functionsToMoveInEnclave: dict of functions with names, also ecall"""

	logging.debug("searching ecall called functions")
	functionsNamesToMoveInEnclave = set()  # the set of funcs names that need to be moved in enclave, that are not ecall

	# find all functions that need to be trusted code
	librariesToInclude = set()  # the libraries that must be included in enclave
	for ecallName, ecall in ecallFunctions.items():  # for each ecall
		ecallReachableEdges = ecall.getFunctionCalledEdgesRecursive()  # get the edges reachable: ret [{source: functName1, dest: funcName2}]

		# init of algorithm
		seen = set()
		todo = set()
		todo.add(ecallName)

		# algorith of graph exploring:  breadth-first search, BFS
		while todo:
			now = todo.pop()
			seen.add(now)
			todo.update(edge["dest"] for edge in ecallReachableEdges if edge["source"] == now)
			todo -= seen
			todo -= ocallFunctions.keys()  # explore only the function not used by ocalls

		functionsNamesToMoveInEnclave.update(seen)  # the list "seen" represent the functions used by the ecall
		logging.debug(f"""the functions {seen} are used by ecall {ecallName}, they will be moved to trusted code""")

		# find the libraries where these functions are defined
		libraryFunctions = seen - allFunctions.keys()  # these are the functions used by this ecall, not defined by the user,
		libraries = set(gccInterface.getHeadersWhereFunctionsAreDef(libraryFunctions,
		                                                            filesPaths).values())  # these are the libraries where these functions can be found
		librariesToInclude.update(libraries)

		logging.debug(
			f"""the standard library functions {libraryFunctions} are used by ecall {ecallName}, {libraries} are included""")

		# the set contains only the set of functions user def used by ecalls that are not ecalls
		functionsNamesToMoveInEnclave -= libraryFunctions  # keep only user def functions in the set

	functionsToMoveInEnclave = {}
	for functionNameToMove in functionsNamesToMoveInEnclave:
		functionsToMoveInEnclave[functionNameToMove] = allFunctions[functionNameToMove]

	logging.info(f"""the libraries used by ecalls are {librariesToInclude}, they will be included""")

	return librariesToInclude, functionsToMoveInEnclave


def checkTheseNeverCallThose(these: {}, those: {}):
	print("\n\ncheck result")
	for this in tqdm(these.values(), leave=True,bar_format='{l_bar}{bar}| time: {elapsed}'):  # this is for the loading bar
		usedBythis = this.getCalledFunctioNames(recursive=False)  # get functions names used directly this code
		for that in usedBythis:
			if that in those and that != "main":
				raise Exception(f"""Error #B003: the untrusted code and trusted code use a same function,
				The function {this} can't use the function {that}. """)


def checkSeparationOfContext(ecallsFunctions, enclaveFunctions, ocallsFunctions, allFunctions):
	"""check that the enclave functions never call untrusted functions unless ocalls
	and checks that the untrusted functions never calls enclave functions unless ecalls"""

	trustedFunctions = {}  # all trusted functions that are not ecalls
	untrustedFunctions = {}  # all untrusted functions that are not ocalls

	for functionName, function in allFunctions.items():
		if functionName in enclaveFunctions and functionName not in ecallsFunctions:  # is trusted but no ecall
			trustedFunctions[functionName] = function
		elif functionName not in enclaveFunctions and functionName not in ocallsFunctions:  # is untrusted but not ocall
			untrustedFunctions[functionName] = function

	# check untrusted functions never call trusted functions unless ecalls
	logging.debug("check trusted function never calls untrusted ones")
	checkTheseNeverCallThose(these={**untrustedFunctions, **ocallsFunctions}, those=trustedFunctions)

	# check enclave functions never call untrusted functions unless ocalls
	logging.debug("check untrusted function never calls trusted ones")
	checkTheseNeverCallThose(these={**enclaveFunctions, **ecallsFunctions}, those=untrustedFunctions)

	logging.info(f"""ok, the functions inside the enclave are {list(trustedFunctions.keys()) + list(ecallsFunctions.keys())}, they will be moved""")
	logging.info(f"""ok, the functions outside the enclave are {list(untrustedFunctions.keys()) + list(ocallsFunctions.keys())}, they are kept""")


def checkArgumentsTypeDefinitionInFunctions(structs, enums, unions, functions: {}):
	"""check ecall arguments type values.
	functions argument is an array of object Function.
	check if the definitions of struct, enum, unions exist"""

	for function in functions.values():  # for each function
		logging.debug(f"""searching arguments definition of {function.functionName}""")
		for arg in function.getArguments():  # for each argument
			argType = arg["argType"]
			typeName = arg["typeName"]
			pointers = arg["pointers"]
			argName = arg["argName"]

			if argType != "simple":  # check if the user def type is well-defined
				found = False
				if argType == "struct":
					if typeName in structs:
						found = True

				elif argType == "enum":
					if typeName in enums:
						found = True

				elif argType == "union":
					if typeName in unions:
						found = True

				if not found:
					raise Exception(f"""Error #B004:  {argType} {typeName} not found in files """)
				else:
					logging.debug(f"""definition ok: {argType} {typeName} """)


def assertDifferentNamesOcallEcall(ecallFunctions: {}, ocallFunctions: {}):
	sameFunctionNames = set(ecallFunctions.keys()).intersection(
		set(ocallFunctions.keys()))  # the ecall and ocall with same name
	if sameFunctionNames:
		raise Exception("Error #B005: ocall and ecall must have different names!" + str(sameFunctionNames))
	else:
		logging.debug("check unique names: ok!")


def copyRemoveAndCopyTree(filepath : list ,src: str, dst: str = "generated_trusted" ):
	userDirName: str = "application"
	userDir = f"{dst}/{userDirName}"

	logging.info(f"Copying all source three in: {userDir}")
	try:
		if os.path.exists(dst) and os.path.isdir(dst):
			shutil.rmtree(dst)

		shutil.copytree(src, userDir)
		os.chdir(userDir)
		filelist = [ f for f in os.listdir("../../" + src) if f ]
		for file in filelist:
			if file not in filepath:
				os.remove(os.path.join(".",file))
		logging.info(f"Working on: {userDir}")

	except Exception as e:
		logging.error(e)
		raise e


def checkFilesExistence(filesPath: []):
	"""Check the existence of each file"""
	logging.debug("Check files existence")
	for filePath in filesPath:
		if os.path.exists(filePath) and os.path.isfile(filePath):
			logging.debug(f"""file: {filePath} exists""")
		else:
			raise Exception(f"""Error #B006: file {filePath} not found""")


def checkAllFileCOrH(filePaths: []):
	""" Check file extension"""
	logging.debug("Check file extension")
	for filePath in filePaths:
		if not filePath.lower().endswith(('.c', '.h')):
			logging.error(f"""Unknown extension for file {filePath}""")


def checkToolsAvailable():
	"""Return an exception if the tool are not available in console"""
	ctagsInterface.isToolAvailable()
	framaInterface.isToolAvailable()
	gccInterface.isToolAvailable()


def substituteNameOfSourceFile(filePath: str):
	"""Replaces the stub name of the application in the makefile with the name of the source file converted"""
	with open("../Makefile", "r") as t:
		filedata = t.read()
		filedata = filedata.replace("#app_name#", filePath).replace("[[SOURCE_NAME]]", filePath.split(".")[0])
		t.close()
	with open("../Makefile", "w") as w:
		w.write(filedata)
		w.close()
	with open("../autorun.py", "r") as p:
		filedata = p.read()
		filedata = filedata.replace("#app_name#", filePath.split(".")[0])
		p.close()
	with open("../autorun.py", "w") as pw:
		pw.write(filedata)
		pw.close()


def checkServerArgumentsAvailable(signerSPID: str, iasPrimaryKey: str, iasSecondaryKey: str, iasReportFilePath: str):
	""" Checks that the arguments of the server read from frameworkConfig are valid"""
	# first checks on signerSPID
	if not signerSPID or not iasPrimaryKey or not iasSecondaryKey or not iasReportFilePath:
		raise Exception("Error #G006: parameters for server not inserted in frameworkConfig")
	if len(signerSPID) < 32:
		raise Exception("Error #G007: SPID must be 32-byte hex string.")
	if len(iasPrimaryKey) < 32:  # the length of the Primary Subscription Key must be the same of the IAS Subscription Key Size(32, as in iasrequest.h
		raise Exception("Error #G008: IAS Primary Subscription Key must be 32-byte hex string.")
	if len(iasSecondaryKey) < 32:
		raise Exception("Error #G009: IAS Primary Subscription Key must be 32-byte hex string.")


def checkPragmaCodeSplitting(cFile: cFileReaderWriter):
	""" Scan the code in order to find #PRAGMA MOVE_TO_SERVER notes """
	logging.debug("Search for #PRAGMA MOVE_TO_SERVER notes")

	requests = regex.findall(r"#\s*pragma\s*move_to_server_E\d+\n", cFile.contents, flags=regex.MULTILINE)
	if(requests.__len__() == 0):
		requests = regex.findall(r"#\s*pragma\s*move_to_server[^\_E\d+]", cFile.contents, flags=regex.MULTILINE)
		return int(requests.__len__())
	else:
		return int(requests.__len__())


def moveFunctionsToServer(cFile: cFileReaderWriter, filePath, remote_attestation: int,portServerCS:int,enclaveToAttest):
	""" It finds the functions that must be moved to the server, substitute the functions and generates the server """
	pragmas = regex.findall(r"(#\s*pragma\s*move_to_server\d*\s*\n\w+\s+\w+)", cFile.contents, flags=regex.MULTILINE)
	#pragmas = regex.findall(r"(#\s*pragma\s*move_to_server_E\d*\s*\n\w+\s+\w+)", cFile.contents, flags=regex.MULTILINE)
	functions = []
	num_arguments = []
	if not pragmas:
		return pragmas
	for f in pragmas:
		nameFunctionRequested = f.split('\n')[1].lstrip().split(' ')[1]
		for nameFunctionClient in cFile.functions:
			if cFile.functions[nameFunctionClient].functionName == nameFunctionRequested:
				# function has the pragma CA
				functions.append(cFile.functions[nameFunctionClient])
				break
	id_f = 0  # this is the id of the function: it will be substituted in the body of the new function that will now call call_remote_function

	for f in functions:
		newFunc = Function("", f.returnType, f.functionName, f.arguments, f.implementationCode, f.entireCode, filePath)
		arguments = newFunc.arguments
		arguments = arguments.replace("(", "").replace(")", "").split(",")
		params = ""
		types_arguments = []
		for a in arguments:
			a = a.lstrip()  # remove all initial spaces
			if params != "":
				params += ","
			if(a.split(" ")[0] == "unsigned"):
				types_arguments.append("unsigned_int")
				a = a.split(" ")[2]
			else:
				types_arguments.append(a.split(" ")[0])
				a = a.split(" ")[1]
			params += a
		num_arguments.append(int(arguments.__len__()))
		# Check that the types of the parameters of the function are basic types
		res_check = CodeSplittingInterface.checkTypeParams(types_arguments)
		if res_check < 0:
			raise Exception(
				f"Error #H001: Parameters of the function {newFunc.functionName} aren't basic data types(functions moved to server must be self contained)")
		# Check that in the code global variables are not used
		res_check = CodeSplittingInterface.checkNoGlobalVars(newFunc.implementationCode, params, filePath)
		if res_check < 0:
			raise Exception(
				f"Error #H002: The function {newFunc.functionName} must not use global variables(functions moved to the server must be self contained")
		# Check that parameters of the function are called by value and not by reference
		res_check = CodeSplittingInterface.checkVarsCopiedByValue(arguments)
		if res_check < 0:
			raise Exception(
				f"Error #H003: Parameters of the function {newFunc.functionName} must be copied by value and not by reference(functions moved to server must be self contained")

		newFunc.implementationCode = "{" + f"""
		function_t *{f.functionName} = (function_t *)malloc(sizeof(function_t));
		{f.functionName}->function_id = {id_f};
		{f.functionName}->num_arguments = {num_arguments[id_f]};
		{f.functionName}->return_type = '{f.returnType[0]}';
		{f.functionName}->return_value.type = '{f.returnType[0]}';
		param_t vector[{f.functionName}->num_arguments];"""

		# returnTypeParam returns ["f",value_f] = [X,Y] where X is type value and Y is field for that type
		for i in range(num_arguments[id_f]):
			newFunc.implementationCode += f"""
		param_t _op{i + 1};
		_op{i + 1}.type = '{CodeSplittingInterface.returnTypeParam(types_arguments[i])[0]}';              
		_op{i + 1}.{CodeSplittingInterface.returnTypeParam(types_arguments[i])[1]} = {params.split(",")[i]};
		vector[{i}] = _op{i + 1};
		"""

		newFunc.implementationCode += f"""
		{f.functionName}->arguments = vector; 
		
		if(call_remote_function({f.functionName})!=0)""" + r"""{
			perror("Error during connection with the server\n");
			abort();
		}""" + f"""
		return {f.functionName}->return_value.[[VALUE_SUBSTITUTE]]; 
{"}"}"""  # f-strings don't allow single brackets
		if newFunc.returnType == "char" or newFunc.returnType == "char*": # or vect[i].lstrip().split(" ")[1].lstrip().startswith("*"):
			newFunc.implementationCode = newFunc.implementationCode.replace("[[VALUE_SUBSTITUTE]]", "buf")
		elif newFunc.returnType == "int":
			newFunc.implementationCode = newFunc.implementationCode.replace("[[VALUE_SUBSTITUTE]]", "value")
		elif newFunc.returnType == "float":
			newFunc.implementationCode = newFunc.implementationCode.replace("[[VALUE_SUBSTITUTE]]", "value_f")
		elif newFunc.returnType == "unsigned_int":
			newFunc.implementationCode = newFunc.implementationCode.replace("[[VALUE_SUBSTITUTE]]", "value_u")
		elif newFunc.returnType == "double":
			newFunc.implementationCode = newFunc.implementationCode.replace("[[VALUE_SUBSTITUTE]]", "value_d")

		newFunc.entireCode = f"""{newFunc.returnType} {newFunc.functionName} {newFunc.arguments} {newFunc.implementationCode}"""
		cFile.replaceFunction(f, newFunc)
		id_f = id_f + 1
	logging.debug("The functions have been replaced with the call to call_remote_function")

	# now we need to insert the call_remote_function implementation
	crfFunction = Function("", "int", "call_remote_function", "(function_t* f)",
	                       CodeSplittingInterface.__crfImplementationCode__, CodeSplittingInterface.__crfEntireCode__,
	                       filePath)
	if enclaveToAttest.__len__()>1:
		string = str(enclaveToAttest[0])
	else:
		string = str(enclaveToAttest)
	crfFunction.implementationCode = crfFunction.implementationCode.replace("attest_enclave_[i]",f"attest_enclave_{string}")
	crfFunction.entireCode = crfFunction.entireCode.replace("attest_enclave_[i]",f"attest_enclave_{string}")
	cFile.addFunction(crfFunction)
	# insert of the struct param_t and struct function
	#requests = regex.findall(r"(#\s*define\s+[\w]+.+)", cFile.contents, flags=regex.MULTILINE)
	requests = regex.findall(r"(int call_remote_function\()", cFile.contents, flags=regex.MULTILINE) #COMMENT IF IT DOESN'T WORK
	cFile.addCodeToTheStart(r"""
typedef struct{
	char type; // i = int f = float d = double c = char u = unsigned int
	int value;
	float value_f;
	char* buf;
	double value_d;
	unsigned int value_u;
} param_t;
typedef struct{
	int function_id;
	int num_arguments;
	param_t *arguments;
	char return_type;
	param_t return_value;
} function_t;""")
	cFile.addInclude("arpa/inet.h")  # must for inet_addr()
	if str(portServerCS) != "":
		cFile.contents = cFile.contents.replace("@REPLACE_DEFAULT_PORT@",f"{str(portServerCS)}")
	# generation of the server
	with open("../server_cs.c", "w") as file:
		logging.info("Writing server_cs.c")
		# we modify the serverCSEntireCode in this section( we need to do it for each function)
		serverCode = CodeSplittingInterface.__serverCSEntireCode__
		code_functions_server = ""
		switch_case_functions = ""
		for i in range(id_f):
			code_functions_server += f"{functions[i].entireCode} \n"
			switch_case_functions += f"case {i}:" + '{' + f"\n\t\t{CodeSplittingInterface.returnTypeResultFunction(functions[i].returnType)} = {functions[i].functionName}{CodeSplittingInterface.returnFunctionArgsCRF(functions[i].arguments,num_arguments[i])};\n\t\t"
			switch_case_functions += f"""sprintf(buff,"{CodeSplittingInterface.returnFormatSpecifiers(functions[i].returnType)}",{CodeSplittingInterface.returnTypeResultFunction(functions[i].returnType)});\n\t\twrite(new_socket_fd, buff, sizeof(buff));\n\t\tbreak;""" + "\n\t\t}\n\t"

		serverCode = serverCode.replace("[SWITCH_CASE_FUNCTIONS]", switch_case_functions).replace("[CODE_FUNCTIONS]",
		                                                                                          code_functions_server)
		serverCode = serverCode.replace("{MAX_PARAMETERS_STUB}",str(max(num_arguments)))
		if str(portServerCS) != "":
			serverCode = serverCode.replace("@REPLACE_DEFAULT_PORT@", f"{str(portServerCS)}")
		# code modified, we can save it,
		file.write(serverCode)
		file.close()
	shutil.copytree("../../include/materialRA/server", "../server_material")  # server_cs needs the libraries, so we move them after having generated the source code



def allowSkipMREnclave():
	""" DEBUG ONLY, allows the skip of MRENCLAVE check from enclave_verify.c in order to show off the RA process"""
	if not os.path.isfile("../server_material/enclave_verify.h"):
		return
	with open("../server_material/enclave_verify.h","r+") as file:
		text = file.read()
		text = text.replace("0","1")
		file.close()
	with open("../server_material/enclave_verify.h", "r+") as file:
		file.write(text)
		file.close()

def checkSDKDir(installdir:str):
	""" Check that the Intel SGX SDK Dir specified in the config file exists"""
	if installdir.startswith('~'):
		installdir = '/home/' + os.path.abspath(installdir).split("/")[2] + f"""/{installdir.split("/")[1]}"""
	if not os.path.exists(installdir):
		raise Exception(f"Error #G010:The directory of installation of the Intel SGX SDK doesn't exists.")

def changeInstallDir(installdir:str):
	"""Change in the makefile the parameter SGX_SDK according to the user configuration in the config file"""
	with open("../Makefile","r") as makefile:
		content_file = makefile.read()
		content_file = content_file.replace("~/sgxsdk",installdir)
		makefile.close()
	with open("../Makefile","w") as makefile:
		makefile.write(content_file)
		makefile.close()

def updateMakefileRA(enclaves_requested:int):
	""" When a server for code splitting is generated, we insert the enclave_u.objects needed to compile the server_cs application """
	with open("../Makefile","r") as makefile:
		content_file = makefile.read()
		enclaves_needed = ""
		for it in range(1,enclaves_requested+1):
			enclaves_needed += f"application/untrusted{str(it)}/enclave{str(it)}_u.o "

		content_file = content_file.replace("@replace_enclave_needed@",enclaves_needed)
		content_file = content_file.replace("$(MAKE) target","$(MAKE) target server_cs").replace("######## Enclave Objects ########",CodeSplittingInterface.__server_compilation_makefile_code__)
		makefile.close()

	with open("../Makefile","w") as makefile:
		makefile.write(content_file)
		makefile.close()

def includeCustomReaction(customReaction:tuple):
	""" If a custom reaction function is asked, it's included in the project"""
	shutil.copy2(f"../../python/addon_files/{customReaction[0]}.c","../server_material")
	if os.path.exists(f"../../python/addon_files/{customReaction[0]}.h"):
		shutil.copy2(f"../../python/addon_files/{customReaction[0]}.h", "../server_material")

	with open("../Makefile","r") as makefile:
		content = makefile.read()
		content = content.replace("$(Enclave_Files)",f"$(Enclave_Files) server_material/{customReaction[0]}.o ")
		content = content.replace(f"server_material/server_interface.cpp: server_material/server_interface.h","server_material/server_interface.cpp: server_material/server_interface.h " +
		                          f"""\n\nserver_material/{customReaction[0]}.o:server_material/{customReaction[0]}.c\n\t@$(CXX) -g $(App_Include_Paths) -c $< -o $@\n\t@echo "GEN => $@" """)
		makefile.close()
	with open("../Makefile","w") as makefile:
		makefile.write(content)
		makefile.close()
	with open("../server_material/server_interface.h","r") as m:
		contentS = m.read()
		contentS = contentS.replace("<vector>",f"""<vector>\n#include "{customReaction[0]}.h" """)
		m.close()
	with open("../server_material/server_interface.h", "w") as m:
		m.write(contentS)
		m.close()
	with open("../server_material/server_interface.cpp", "r") as cw:
		code = cw.read()
		code = code.replace(f"""eprintf("error processing msg3\n");""",f"""eprintf("error processing msg3\n");\n\t"""+"helper_function();")
		cw.close()
	with open("../server_material/server_interface.cpp","w") as cw:
		cw.write(code)
		cw.close()

def search_enclaves_needing_attestation(cFileReaderWriterDict: cFileReaderWriter):
	""" Search the enclaves to attest and returns their id"""
	requests = regex.findall(r"(#\s*pragma\s*attest_enclave_\d+)", cFileReaderWriterDict.contents, flags=regex.MULTILINE)
	enclaves = []
	if requests:
		for r in requests:
			id_enclave = r.split("enclave_")[1]
			if id_enclave not in enclaves:
				enclaves.append(id_enclave)
	return enclaves
